/*
30-June-2005  Drew Baden
1. Start with s2 version of the sandwich self-triggering board....
2. Modify so that each Xilinx in the HTR sends, at 80MHz
    o 2 bits per channel (24 channels, 48 bits)
	 o 16 bits for sumET
	 o 1 bit for "frame"
	 o 1 bit for L1A
  Total: 66 bits @ 80MHz
3. We need some output pins to debug, so we'll use the "Wesley" connectors:
  J4 and J5 are the 50pin connectors that are available and look like this
  on the board:
  
     J4          J5         J7
  -50- -25-   -50- -25-    |||||
  -49- -24-   -49- -24-    |||||
  
  -27- -02-   -27- -02-    |||||
  -26- -01-   -26- -01-    |||||
                            J6
on J4 and J5:  pin 1 is the bottom right pin in the 25row by 2 columns

On J4, there are 23 free pins
On J5, there are 23 free pins

Pins 49 and 50 on J4 are REFCLK+ and REFCLK- driven from the FPGA
on B12.  The firmware below drives this pin (B12) from the DCM through
a TTL->LVPECL converter (U2A).  The origin of this clock is the 40MHz
clock from the TTCrx

We need an 80MHz clock to latch things.  The DCM produces a 80MHz clock
  which is 2x the TTC clock.  Drive this onto J5/pin1 from U3 (CK80)

Data from each HTR Xilinx:

*/  

`include "define.v"
module SandwichTop 
 ( 
 input  CLKp, CLKn, // 40MHz differential RX_CLK
 input  BCOp, BCOn, // RX_BC0 common to all SLB boards
 input  SLB_ID,
 input  TTC_CK40Des1, TTC_BRCST_STRB1, //TTC_BCR,
 input  SCLK,       // 80 MHz Clock from onboard crystal
 input  LOC_nCS,  LocWRITE_b, 
 input  [7:2] TTC_BRCST,
 input [23:0] TOP_D, BOT_D,		// data from HTR motherboard
 input [11:0] TOPbus, BOTbus,    //               "
 input  [`LB_Addr_Size-1:0]  LocAdd,  // The Address bus.
 inout  [`REG_WIDTH-1:0]     LocBus,  // The bidir Data bus 
 output SLB_CFGDONE,
 output [22:0] SPY4,		// ouput onto J4  see http://www.physics.umd.edu/hep/HTR/luminosity/uwspy.html
 output [22:0] SPY5,		// ouput onto J5                            ditto

 output LVDS1n, LVDS1p, LVDS2n, LVDS2p,// to MDR14 J10
 output REFCLK, TRSTN,  LED_CONTROL          );

 wire  VME_Hard_rst, VME_Soft_rst, VME_Start, VME_Stop, VME_LVDSout,
       run_status, BunchEr, GateTTCcommands, Hard_rst, Soft_rst;
 wire [15:0] TTC_test0, TTC_test1, TTC_test2, TTC_test3,
             TTC_test4, TTC_test5, TTC_test6, TTC_test7,
             TTC_test8, TTC_test9, TTC_testA, TTC_testB,
             TTC_testC, TTC_testD, TTC_testE, TTC_testF;

 wire [8:1]  TPSpyFifoEmpty, TPSpyFifoFull, TPSpyFifoRd,
             InstantCntErr,  InstantR1sErr;
 wire [15:0] AccCntErr, AccR1sErr;  
 wire [35:0]  TPSpyFIFO1_out, TPSpyFIFO2_out, TPSpyFIFO3_out;
 reg VME_LVDSout_D, LVDSout1, LVDSout2;
 wire [7:0] LastTTCcommand;
 reg [23:0] rTOP_D, rBOT_D;
 reg [11:0] rTOPbus, rBOTbus;
 reg [15:0] RxClkCnt, RxBC0Cnt, TClkCnt;
 wire[7:0] mux_out;

//  ********************************************************************
//      ********************  CLOCK SECTION    *****************
//            *********************************************

 wire   RxClk_Buf, SCLK_buf;
 wire   Tclk0, TTC_CK, Tclk2x_dll, TTC_CK2x, TTC_DLL_lock;
 reg    [2:0] UnlockCnt; // counter for proper reset of DLL 

 // Instantiate the Differential Global Input Buffer 
  IBUFGDS_LVPECL_25 difclkbuf (.I(CLKp), .IB(CLKn), .O(RxClk_Buf));
  IBUFG sclkbuf ( .I(SCLK), .O(SCLK_buf) );
                  
// Generate a DLL reset (3-ticks) after FPGA is configured, see:
// www.xilinx.com/publications/xcellonline/xcell_51/xc_particle51.htm
  FDS flop1 (.D(1'b0), .C(SCLK_buf), .Q(out1), .S(1'b0)); 
  FD  flop2 (.D(out1), .C(SCLK_buf), .Q(out2)); 
  FD  flop3 (.D(out2), .C(SCLK_buf), .Q(out3)); 
  FD  flop4 (.D(out3), .C(SCLK_buf), .Q(out4)); 
  assign ConfigRst = (out2 | out3 | out4); 

  DCM TTC_DCM
   (.CLKIN(TTC_CK40Des1), .CLKFB(TTC_CK),  .RST(ConfigRst),  
    .CLK0(Tclk0),   .CLK90(),  .CLK180(),  .CLK270(),	
    .CLK2X(),  .CLK2X180(Tclk2x_dll),  // here u can use the inverted clock.
    .CLKDV(),   .CLKFX(),   .CLKFX180(),
    .LOCKED(TTC_DLL_lock),  .STATUS(),
    .DSSEN(1'b0),
    .PSCLK(1'b0),  .PSEN(1'b0),  .PSINCDEC(1'b0), .PSDONE() );

  BUFG TCK40BUF (.I(Tclk0), .O(TTC_CK));  
  BUFG TCK2XBUF (.I(Tclk2x_dll), .O(TTC_CK2x)); 


//       *************  END OF CLOCK SECTION  *****************
//  *****************************************************************
// **** Register, Spy and Check the HTR data ***

//
//register data since it is arriving at 80MHz
//
wire frame = TOPbus[11];
 always @ (posedge TTC_CK2x) begin 
  rTOP_D <= TOP_D;
  rBOT_D <= BOT_D;
  rTOPbus <= TOPbus;
  rBOTbus <= BOTbus;
  end
reg[26:0] clk_scaled;
//
// now form the 2-bit data for each channel
//
wire rframe = rTOPbus[11];  
wire rTTC_TestEnable = rTOPbus[10];
wire[1:0] ch1,ch2,ch3;  //TOP fiber 1
wire[3:0] mch1,mch2,mch3;  //TOP fiber 1
htrchannel CH1 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[0]),
	.chlevel(ch1), .chdemux(mch1) );
htrchannel CH2 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[1]),
	.chlevel(ch2), .chdemux(mch2) );
htrchannel CH3 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[2]),
	.chlevel(ch3), .chdemux(mch3) );

wire[1:0] ch4,ch5,ch6;  //TOP fiber 2
wire[3:0] mch4,mch5,mch6;  //TOP fiber 2
htrchannel CH4 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[3]),
	.chlevel(ch4), .chdemux(mch4) );
htrchannel CH5 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[4]),
	.chlevel(ch5), .chdemux(mch5) );
htrchannel CH6 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[5]),
	.chlevel(ch6), .chdemux(mch6) );

wire[1:0] ch7,ch8,ch9;  //TOP fiber 3
wire[3:0] mch7,mch8,mch9;  //TOP fiber 3
htrchannel CH7 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[6]),
	.chlevel(ch7), .chdemux(mch7) );
htrchannel CH8 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[7]),
	.chlevel(ch8), .chdemux(mch8) );
htrchannel CH9 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[8]),
	.chlevel(ch9), .chdemux(mch9) );

wire[1:0] ch10,ch11,ch12;  //TOP fiber 4
wire[3:0] mch10,mch11,mch12;  //TOP fiber 4
htrchannel CH10 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[9]),
	.chlevel(ch10), .chdemux(mch10) );
htrchannel CH11 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[10]),
	.chlevel(ch11), .chdemux(mch11) );
htrchannel CH12 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[11]),
	.chlevel(ch12), .chdemux(mch12) );

wire[1:0] ch13,ch14,ch15;  //TOP fiber 5
wire[3:0] mch13,mch14,mch15;  //TOP fiber 5
htrchannel CH13 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[12]),
	.chlevel(ch13), .chdemux(mch13) );
htrchannel CH14 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[13]),
	.chlevel(ch14), .chdemux(mch14) );
htrchannel CH15 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[14]),
	.chlevel(ch15), .chdemux(mch15) );

wire[1:0] ch16,ch17,ch18;  //TOP fiber 6
wire[3:0] mch16,mch17,mch18;  //TOP fiber 6
htrchannel CH16 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[15]),
	.chlevel(ch16), .chdemux(mch16) );
htrchannel CH17 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[16]),
	.chlevel(ch17), .chdemux(mch17) );
htrchannel CH18 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[17]),
	.chlevel(ch18), .chdemux(mch18) );

wire[1:0] ch19,ch20,ch21;  //TOP fiber 7
wire[3:0] mch19,mch20,mch21;  //TOP fiber 7
htrchannel CH19 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[18]),
	.chlevel(ch19), .chdemux(mch19) );
htrchannel CH20 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[19]),
	.chlevel(ch20), .chdemux(mch20) );
htrchannel CH21 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[20]),
	.chlevel(ch21), .chdemux(mch21) );

wire[1:0] ch22,ch23,ch24;  //TOP fiber 8
wire[3:0] mch22,mch23,mch24;  //TOP fiber 8
htrchannel CH22 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[21]),
	.chlevel(ch22), .chdemux(mch22) );
htrchannel CH23 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[22]),
	.chlevel(ch23), .chdemux(mch23) );
htrchannel CH24 (.clk(TTC_CK), .clk2x(TTC_CK2x), .frame(rframe), .chbit(rTOP_D[23]),
	.chlevel(ch24), .chdemux(mch24) );
//
// for histograms, count the number of channels in each individual threshold
// this number can go from 0 to 48, so we need a number which has 6 bits (0-63)
// (STILL HAVE TO ADD THE BOTTOM!)
//
reg[5:0] sum0, sum1, sum2, sum3;
always @ (posedge TTC_CK) begin
  sum0 = mch1[0] + mch2[0] + mch3[0] + mch4[0] + mch5[0] + mch6[0] +
         mch7[0] + mch8[0] + mch9[0] + mch10[0] + mch11[0] + mch12[0] +
         mch13[0] + mch14[0] + mch15[0] + mch16[0] + mch17[0] + mch18[0] +
         mch19[0] + mch20[0] + mch21[0] + mch22[0] + mch23[0] + mch24[0];
  sum1 = mch1[1] + mch2[1] + mch3[1] + mch4[1] + mch5[1] + mch6[1] +
         mch7[1] + mch8[1] + mch9[1] + mch10[1] + mch11[1] + mch12[1] +
         mch13[1] + mch14[1] + mch15[1] + mch16[1] + mch17[1] + mch18[1] +
         mch19[1] + mch20[1] + mch21[1] + mch22[1] + mch23[1] + mch24[1];
  sum2 = mch1[2] + mch2[2] + mch3[2] + mch4[2] + mch5[2] + mch6[2] +
         mch7[2] + mch8[2] + mch9[2] + mch10[2] + mch11[2] + mch12[2] +
         mch13[2] + mch14[2] + mch15[2] + mch16[2] + mch17[2] + mch18[2] +
         mch19[2] + mch20[2] + mch21[2] + mch22[2] + mch23[2] + mch24[2];
  sum3 = mch1[3] + mch2[3] + mch3[3] + mch4[3] + mch5[3] + mch6[3] +
         mch7[3] + mch8[3] + mch9[3] + mch10[3] + mch11[3] + mch12[3] +
         mch13[3] + mch14[3] + mch15[3] + mch16[3] + mch17[3] + mch18[3] +
         mch19[3] + mch20[3] + mch21[3] + mch22[3] + mch23[3] + mch24[3];
  end
//
// spy output onto wisconsin receiver
//	see http://www.physics.umd.edu/hep/HTR/luminosity/uwspy.html
// there are 23 pins available per wisconsin connectors J4 and J5
// SPY4 has is hooked up to various bits of the 16-bit Agilent LA
//                1          1       1          1            6 (use 0)    1	        1             10
assign SPY4 = {TOP_D[23], TTC_CK, TTC_CK2x, TTC_BRCST_STRB1, TTC_BRCST, rframe, rTTC_TestEnable, rTOPbus[9:0]};
//
// SPY5 is hooked up so that the bottom 12 bits of the agilent are in order
//
assign SPY5[22:16] = 'h0;
assign SPY5[15:12] = {TTC_CK2x,TTC_CK,TTC_BRCST_STRB1,rframe};
reg[11:0] spy;
assign SPY5[11:0] = spy;
always @ (mux_out)
	case (mux_out[1:0])
		0: spy[11:0] = {ch2,ch1,mch1,TOP_D[1:0],rTOP_D[1:0]};
		1: spy[11:0] = {ch2,ch1,mch2,TOP_D[1:0],rTOP_D[1:0]};
		2: spy[11:0] = {sum1,sum0};
		3: spy[11:0] = {sum3,sum2};
   endcase

always @ (posedge TTC_CK2x) clk_scaled <= #1 clk_scaled + 1;
assign LED_CONTROL = clk_scaled[25];

//LED_Sine breathing(.clock(SCLK_buf), .nReset(1'b1), .led(LED_CONTROL));
 assign REFCLK = TTC_CK;
 assign SLB_CFGDONE = SLB_ID; // in case ECAL TCC uses this board
 assign TRSTN = 1'b1; // Wisc. Rec_Mezz JTAG reset.
 IBUFDS_LVPECL_25 RxBC0buf (.I(BCOp), .IB(BCOn), .O(RxBC0));
 OBUFDS_LVDS_25 pair1 (.I(LVDSout1), .O(LVDS1p), .OB(LVDS1n));
 OBUFDS_LVDS_25 pair2 (.I(LVDSout2), .O(LVDS2p), .OB(LVDS2n));
//
// decode the TTC to keep track (just for checking)
//
decodeTTC TTCdecode (
 .TTC_CK40Des1(TTC_CK),
 .reset(VME_Hard_rst), 
 .TTC_Brcst(TTC_BRCST),
 .TTC_BrcstStr(TTC_BRCST_STRB1),   
 .TTC_test0(TTC_test0), .TTC_test1(TTC_test1),
 .TTC_test2(TTC_test2), .TTC_test3(TTC_test3),
 .TTC_test4(TTC_test4), .TTC_test5(TTC_test5),
 .TTC_test6(TTC_test6), .TTC_test7(TTC_test7),
 .TTC_test8(TTC_test8), .TTC_test9(TTC_test9),
 .TTC_testA(TTC_testA), .TTC_testB(TTC_testB),
 .TTC_testC(TTC_testC), .TTC_testD(TTC_testD),
 .TTC_testE(TTC_testE), .TTC_testF(TTC_testF),
 .LastTTCcommand(LastTTCcommand)
 );

always @ (posedge TTC_CK) begin
  LVDSout1 <= VME_LVDSout;
  LVDSout2 <= VME_LVDSout;
  end

//
// put the entire 36 bits from each Xilinx (HTR TOP and BOT) into a fifo
// which is 512 deep and 36 bits wide (RAMB16_S36_S36).  however, the
// readout data path is only 16 bits wide.  what to do?   we'll have to 
// read it 16 bits at a time!	 therefore we'll need a different register
// for the 3 reads (3x16 > 36 bits).  all this will be done inside the
// localbusslave routine
//
fifo_ic_v2  TPSpyFIFO1 
 ( .fifo_gsr_in(VME_Hard_rst),       .write_data_in({rTOPbus, rTOP_D}), 
   .write_clock_in(TTC_CK2x),        .write_enable_in(run_status),
   .read_clock_in(TTC_CK),           .read_enable_in(TPSpyFifoRd[1]),
   .read_data_out({TPSpyFIFO1_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[1]),      .empty_out(TPSpyFifoEmpty[1])  );

fifo_ic_v2  TPSpyFIFO2 
 ( .fifo_gsr_in(VME_Hard_rst),       .write_data_in({rBOTbus, rBOT_D}), 
   .write_clock_in(TTC_CK2x),        .write_enable_in(run_status),
   .read_clock_in(TTC_CK),           .read_enable_in(TPSpyFifoRd[2]),
   .read_data_out({TPSpyFIFO2_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[2]),      .empty_out(TPSpyFifoEmpty[2])  );
												 
fifo_ic_v2  TPSpyFIFO3 
 ( .fifo_gsr_in(VME_Hard_rst),       .write_data_in({12'b0,sum3,sum2,sum1,sum0}), 
   .write_clock_in(TTC_CK),        .write_enable_in(run_status),
   .read_clock_in(TTC_CK),           .read_enable_in(TPSpyFifoRd[3]),
   .read_data_out({TPSpyFIFO3_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[3]),      .empty_out(TPSpyFifoEmpty[3])  );
												 
////////////////////////////////////////////////////////


LocBusSlave LocBusSlave (
    .clk(TTC_CK),      .rst(1'b0), 
    .LocAddr(LocAdd),    .LocData(LocBus), 
    .LocCS_b(LOC_nCS),   .LocWRITE_b(LocWRITE_b), 
    .LB_STATE(),   .Din(), .Dout(),      .Tri(), 
    .RxDLL_lock(1'b0), .TTC_DLL_lock(TTC_DLL_lock), .UnlockCnt(UnlockCnt),
    .DLL_status(16'b0),
    .LastTTCcommand(LastTTCcommand), .GateTTCcommands(GateTTCcommands), 
    .VME_Hard_rst(VME_Hard_rst),     .VME_Soft_rst(VME_Soft_rst), 
    .VME_Start(VME_Start),           .VME_Stop(VME_Stop), 
    .VME_LVDSout(VME_LVDSout), 
    .run_status(run_status),         
    .RxClkCnt(RxClkCnt), .RxBC0Cnt(RxBC0Cnt),
    .TClkCnt(TClkCnt), .TxBC0Cnt(16'b0),
    .TPSpyFifoRd(TPSpyFifoRd),
    .TPSpyFifoEmpty(TPSpyFifoEmpty),   .TPSpyFifoFull(TPSpyFifoFull), 
    .TPSpyFIFO1_out(TPSpyFIFO1_out),   .TPSpyFIFO2_out(TPSpyFIFO2_out),
    .TPSpyFIFO3_out(TPSpyFIFO3_out),//   .TPSpyFIFO4_out(TPSpyFIFO4_out),
    .TTC_test0(TTC_test0),       .TTC_test1(TTC_test1), 
    .TTC_test2(TTC_test2),       .TTC_test3(TTC_test3), 
    .TTC_test4(TTC_test4),       .TTC_test5(TTC_test5), 
    .TTC_test6(TTC_test6),       .TTC_test7(TTC_test7), 
    .TTC_test8(TTC_test8),       .TTC_test9(TTC_test9), 
    .TTC_testA(TTC_testA),       .TTC_testB(TTC_testB), 
    .TTC_testC(TTC_testC),       .TTC_testD(TTC_testD), 
    .TTC_testE(TTC_testE),       .TTC_testF(TTC_testF), 
    .BunchEr(BunchEr), .mux_out(mux_out)
        );

 commander commander (  
  .clk(SCLK_buf),    .GateTTCcommands(GateTTCcommands),
  .TTC_Hard_rst(1'b0), .TTC_Soft_rst(1'b0),
  .TTC_Start(1'b0), .TTC_Stop(1'b0), 
  .VME_Hard_rst(VME_Hard_rst), .VME_Soft_rst(VME_Soft_rst), 
  .VME_Start(VME_Start), .VME_Stop(VME_Stop), 
  .Hard_rst(Hard_rst),  .Soft_rst(Soft_rst), 
  .run_status(run_status)       );

 always @ (posedge RxClk_Buf) begin 
   if (Soft_rst) begin
      RxClkCnt <= 0;  RxBC0Cnt <= 0;
   end
   else if (run_status) begin
      RxClkCnt <= RxClkCnt + 1;
      if (RxBC0) RxBC0Cnt <= RxBC0Cnt + 1;
   end
 end // always

 always @ (posedge TTC_CK) begin 
   if (Soft_rst)         TClkCnt <= 0; 
   else if (run_status)  TClkCnt <= TClkCnt + 1;
 end // always

endmodule